Sackville = LibStub("AceAddon-3.0"):NewAddon("Sackville", "AceConsole-3.0", "AceHook-3.0", 'AceEvent-3.0')
local B = LibStub("LibBabble-Inventory-3.0")
local BL = B:GetLookupTable()

-- TODO: collect all options/defaults somewhere
local COLUMN_WIDTH = 40
local ROW_HEIGHT = 40
local BUTTON_HEIGHT = 40
local BUTTON_WIDTH = 40
local FRAME_PADDING = 0


local db = nil
local loading = true
local sprintf = string.format

Sackville.bagIndexes = {}
Sackville.items = {}
Sackville.playerBags = {}
Sackville.bankBags = {}
Sackville.dummyFrame = nil
Sackville.isBankOpen = false
Sackville.bankInitialized = false
Sackville.categories = {
    [BL["Armor"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = "Equipment",
            subCategory = BL["Armor"]
        }
    },
    [BL["Consumable"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = "Consumable"
        }
    },
    [BL["Weapon"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Weapon"]
        }
    },
    [BL["Trinket"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = "Equipment",
            subCategory = BL["Trinket"]
        }
    },
    ["Jewelry"] = {
        [1] = {
            negate = false,
            operation = "or",
            category = "Equipment",
            subCategory = BL["Ring"]
        },
        [2] = {
            negate = false,
            operation = "and",
            category = "Equipment",
            subCategory = BL["Neck"]
        }
    },
    [BL["Tabard"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = "Equipment",
            subCategory = BL["Tabard"]
        }
    },
    [BL["Food & Drink"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Consumable"],
            subCategory = BL["Food & Drink"]
        }
    },
    [BL["Potion"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Consumable"],
            subCategory = BL["Potion"]
        }
    },
    [BL["Elixir"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Consumable"],
            subCategory = BL["Elixir"]
        }
    },
    [BL["Flask"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Consumable"],
            subCategory = BL["Flask"]
        }
    },
    [BL["Scroll"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Consumable"],
            subCategory = BL["Scroll"]
        }
    },
    [BL["Bandage"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Consumable"],
            subCategory = BL["Bandage"]
        }
    },
    [BL["Metal & Stone"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"],
            subCategory = BL["Metal & Stone"]
        }
    },
    [BL["Jewelcrafting"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"],
            subCategory = BL["Jewelcrafting"]
        }
    },
    [BL["Enchanting"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"],
            subCategory = BL["Enchanting"]
        }
    },
    [BL["Cloth"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"],
            subCategory = BL["Cloth"]
        }
    },
    [BL["Parts"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"],
            subCategory = BL["Parts"]
        }
    },
    [BL["Elemental"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"],
            subCategory = BL["Elemental"]
        }
    },
    [BL["Trade Goods"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"]
        }
    },
    [BL["Quest"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Quest"],
            subCategory = BL["Quest"]
        }
    },
    [BL["Reagent"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Miscellaneous"],
            subCategory = BL["Reagent"]
        }
    },
    [BL["Devices"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = BL["Trade Goods"],
            subCategory = BL["Devices"]
        }
    },
    [BL["Junk"]] = {
        [1] = {
            negate = false,
            operation = "and",
            category = "Quality", -- FIXME: localize
            subCategory = ITEM_QUALITY0_DESC
        }
    },
}

Sackville.bagConfig = {
    [1] = {
        name = "Equipment",
        bagType = "player",
        priority = 2,
        sortCriteria = { 
            [1] = {
                key = "level",
                comparison = ">"
            },
            [2] = {
                key = "quality",
                comparison = ">"
            },
            [3] = {
                key = "name",
                comparison = "<"
            }        
        },
        sections = {
            [1] = {
                name = BL["Armor"],
                priority = 1,
                categories = { 
                    BL["Armor"]
                }
            },
            [2] = {
                name = BL["Weapon"],
                categories = { 
                    BL["Weapon"]
                }
            },
            [3] = {
                name = BL["Trinket"],
                categories = { 
                    BL["Trinket"]
                }
            },
            [4] = {
                name = "Jewelry",
                categories = { 
                    "Jewelry"
                }
            },
            [5] = {
                name = BL["Tabard"],
                categories = { 
                    BL["Tabard"]
                }
            }
        }
    },
    [2] = {
        name = BL["Consumable"],
        bagType = "player",
        priority = 2,
        sortCriteria = { 
            [1] = {
                key = "minLevel",
                comparison = ">"
            },
            [2] = {
                key = "name",
                comparison = "<"
            },
            [3] = {
                key = "count",
                comparison = ">"
            }        
        },
        sections = {
            [1] = {
                name = BL["Food & Drink"],
                categories = { 
                    BL["Food & Drink"]
                }
            },
            [2] = {
                name = BL["Flask"],
                categories = { 
                    BL["Flask"]
                }
            },
            [3] = {
                name = BL["Potion"],
                categories = { 
                    BL["Potion"]
                }
            },
            [4] = {
                name = BL["Elixir"],
                categories = { 
                    BL["Elixir"]
                }
            },
            [5] = {
                name = BL["Scroll"],
                categories = { 
                    BL["Scroll"]
                }
            },
            [6] = {
                name = BL["Bandage"],
                categories = { 
                    BL["Bandage"]
                }
            },
            [7] = {
                name = BL["Other"],
                priority = 1,
                categories = { 
                    BL["Consumable"]
                }
            },              
        }
    },
    [3] = {
        name = BL["Trade Goods"],
        bagType = "player",
        priority = 2,
        sortCriteria = { 
            [1] = {
                key = "name",
                comparison = "<"
            },
            [2] = {
                key = "count",
                comparison = ">"
            }        
        },
        sections = {
            [1] = {
                name = BL["Metal & Stone"],
                categories = { 
                    BL["Metal & Stone"]
                }
            },
            [2] = {
                name = BL["Jewelcrafting"],
                categories = {
                    BL["Jewelcrafting"]
                }
            },
            [3] = {
                name = BL["Enchanting"],
                categories = { 
                    BL["Enchanting"]
                }
            },
            [4] = {
                name = BL["Cloth"],
                categories = { 
                    BL["Cloth"]
                }
            },
            [5] = {
                name = BL["Parts"],
                categories = { 
                    BL["Parts"]
                }
            },
            [6] = {
                name = BL["Elemental"],
                categories = { 
                    BL["Elemental"]
                }
            },
            [7] = {
                name = BL["Other"],
                priority = 1,
                categories = { 
                    BL["Elemental"]
                }
            }         
        }
    },
    [4] = {
        name = BL["Other"],
        bagType = "player",
        priority = 1,
        sortCriteria = { 
            [1] = {
                key = "name",
                comparison = "<"
            },
            [2] = {
                key = "count",
                comparison = ">"
            }        
        },
        sections = {
            [1] = {
                name = BL["Quest"],
                categories = { 
                    BL["Quest"]
                }
            },
            [2] = {
                name = BL["Reagent"],
                categories = { 
                    BL["Reagent"]
                }
            },
            [3] = {
                name = BL["Devices"],
                categories = { 
                    BL["Devices"]
                }
            },
            [4] = {
                name = BL["Junk"],
                categories = { 
                    BL["Junk"]
                }
            },
            [5] = {
                name = BL["Other"],
                priority = 0,
                defaultSection = true
            }
        }
    },
    [5] = {
        name = "Bank Equipment",
        bagType = "bank",
        priority = 2,
        sortCriteria = { 
            [1] = {
                key = "level",
                comparison = ">"
            },
            [2] = {
                key = "quality",
                comparison = ">"
            },
            [3] = {
                key = "name",
                comparison = "<"
            }        
        },
        sections = {
            [1] = {
                name = BL["Armor"],
                priority = 1,
                categories = { 
                    BL["Armor"]
                }
            },
            [2] = {
                name = BL["Weapon"],
                categories = { 
                    BL["Weapon"]
                }
            },
            [3] = {
                name = BL["Trinket"],
                categories = { 
                    BL["Trinket"]
                }
            },
            [4] = {
                name = "Jewelry",
                categories = { 
                    "Jewelry"
                }
            },
            [5] = {
                name = BL["Tabard"],
                categories = { 
                    BL["Tabard"]
                }
            }
        }
    },
    [6] = {
        name = "Bank Consumable",
        bagType = "bank",
        priority = 2,
        sortCriteria = { 
            [1] = {
                key = "minLevel",
                comparison = ">"
            },
            [2] = {
                key = "name",
                comparison = "<"
            },
            [3] = {
                key = "count",
                comparison = ">"
            }        
        },
        sections = {
            [1] = {
                name = BL["Food & Drink"],
                categories = { 
                    BL["Food & Drink"]
                }
            },
            [2] = {
                name = BL["Flask"],
                categories = { 
                    BL["Flask"]
                }
            },
            [3] = {
                name = BL["Potion"],
                categories = { 
                    BL["Potion"]
                }
            },
            [4] = {
                name = BL["Elixir"],
                categories = { 
                    BL["Elixir"]
                }
            },
            [5] = {
                name = BL["Scroll"],
                categories = { 
                    BL["Scroll"]
                }
            },
            [6] = {
                name = BL["Bandage"],
                categories = { 
                    BL["Bandage"]
                }
            },
            [7] = {
                name = BL["Other"],
                priority = 1,
                categories = { 
                    BL["Consumable"]
                }
            },              
        }
    },
    [7] = {
        name = "Bank Trade Goods",
        bagType = "bank",
        priority = 2,
        sortCriteria = { 
            [1] = {
                key = "name",
                comparison = "<"
            },
            [2] = {
                key = "count",
                comparison = ">"
            }        
        },
        sections = {
            [1] = {
                name = BL["Metal & Stone"],
                categories = { 
                    BL["Metal & Stone"]
                }
            },
            [2] = {
                name = BL["Jewelcrafting"],
                categories = {
                    BL["Jewelcrafting"]
                }
            },
            [3] = {
                name = BL["Enchanting"],
                categories = { 
                    BL["Enchanting"]
                }
            },
            [4] = {
                name = BL["Cloth"],
                categories = { 
                    BL["Cloth"]
                }
            },
            [5] = {
                name = BL["Parts"],
                categories = { 
                    BL["Parts"]
                }
            },
            [6] = {
                name = BL["Elemental"],
                categories = { 
                    BL["Elemental"]
                }
            },
            [7] = {
                name = BL["Other"],
                priority = 1,
                categories = { 
                    BL["Elemental"]
                }
            }         
        }
    },
    [8] = {
        name = "Bank Other",
        bagType = "bank",
        priority = 1,
        sortCriteria = { 
            [1] = {
                key = "name",
                comparison = "<"
            },
            [2] = {
                key = "count",
                comparison = ">"
            }        
        },
        sections = {
            [1] = {
                name = BL["Quest"],
                categories = { 
                    BL["Quest"]
                }
            },
            [2] = {
                name = BL["Reagent"],
                categories = { 
                    BL["Reagent"]
                }
            },
            [3] = {
                name = BL["Devices"],
                categories = { 
                    BL["Devices"]
                }
            },
            [4] = {
                name = BL["Junk"],
                categories = { 
                    BL["Junk"]
                }
            },
            [5] = {
                name = BL["Other"],
                priority = 0,
                defaultSection = true
            }
        }
    }

}
    
function Sackville:DumpItem(item)

    if item then
        local id = item.id or "?"
        local name = item.name or "?"
        local bagslot = item.bagslot or "?"
        local itemType = item.type or "?"
        local itemSubType = item.subType or "?"
        local removed = item.removed and "true" or "false"
        return sprintf("id: %s, name: %s, bagslot: %s, type: %s, subType: %s, itemRemoved: %s",
            id, name, bagslot, itemType, itemSubType, removed)
    else
        return "nil item"
    end

end

local function PositionBag(vbag, state)
    local bagFrame = vbag:GetFrame()
    local h = bagFrame:GetHeight()
    local w = bagFrame:GetWidth()
    -- Sackville:Print(sprintf(
        -- " PositionBag, x_offset: %d, y_offset: %d, widest_bag: %d, w: %d, h: %d, SH: %d",
        -- state.x_offset, state.y_offset, state.widest_bag, w, h, GetScreenHeight()))    
    state.widest_bag = (w > state.widest_bag) and w or state.widest_bag
    
    if (state.y_offset + h) < GetScreenHeight() then
        -- There is enough room to stack this frame vertically
        -- Sackville:Print(sprintf("  Setting pos to (%d, %d)", state.x_offset, state.y_offset))
        bagFrame:SetPoint("BOTTOMRIGHT", state.x_offset, state.y_offset)
        state.y_offset = state.y_offset + h + FRAME_PADDING
    else
        -- Start a new column
        state.y_offset = 80
        state.x_offset = state.x_offset - state.widest_bag - FRAME_PADDING
        state.widest_bag = 0
        -- Sackville:Print(sprintf("  Setting pos to (%d, %d)", state.x_offset, state.y_offset))
        bagFrame:SetPoint("BOTTOMRIGHT", state.x_offset, state.y_offset)
        state.y_offset = state.y_offset + h + FRAME_PADDING
    end
    
    return state
end

function Sackville:IsBankBagId(bagId)
    for _, bankBagId in ipairs(self.bagIndexes.bank) do
        if bagId == bankBagId then
            return true
        end
    end
    
    return false
end

function Sackville:IsPlayerBagId(bagId)
    for _, playerBagId in ipairs(self.bagIndexes.bags) do
        if bagId == playerBagId then
            return true
        end
    end
    
    return false
end

function Sackville:GetTextDefaults()
    return {
        font = '',
        size = 11,
        flags = {
             ['OUTLINE'] = false,
             ['THICKOUTLINE'] = false,
             ['MONOCHROME'] = false,
             ['SHADOW'] = true,
        },
        color = {1,1,1},
        halign = 'LEFT',
        valign = 'CENTER',
        padding = 1,
    }
end

function Sackville:BuildbagIndexes()
   self.bagIndexes = {
      bags = {},
      bank = {}
   }

   for i = 0, NUM_BAG_SLOTS do
      table.insert(self.bagIndexes.bags, i)
   end
   table.insert(self.bagIndexes.bank, BANK_CONTAINER)
   for i = NUM_BAG_SLOTS + 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do
      table.insert(self.bagIndexes.bank, i)
   end
end

function Sackville:GetMoneyString()
   local copper = GetMoney()
   local money

   local size = 12
   local t = sprintf('%d|TInterface\\MoneyFrame\\UI-GoldIcon:%d:%d:3:0:64:64:4:60:4:60|t ', copper / 100 / 100, size, size)
   t = t..sprintf(' %d|TInterface\\MoneyFrame\\UI-SilverIcon:%d:%d:3:0:64:64:4:60:4:60|t ', (copper / 100) % 100, size, size)
   t = t..sprintf(' %d|TInterface\\MoneyFrame\\UI-CopperIcon:%d:%d:3:0:64:64:4:60:4:60|t ', copper % 100, size, size)

   -- for i=1, GetCurrencyListSize() do
      -- local name, isHeader, _, _, isWatched, count, icon = GetCurrencyListInfo(i)
      -- if (not isHeader) then
         -- if isWatched then
            -- t = (count or 0)..'|T'..icon..':'..size..':'..size..':3:0:64:64:4:60:4:60|t   '..t
         -- end
         -- if money then
            -- money[i] = count
         -- end
      -- end
   -- end

   return t
end

function Sackville:RebuildInventory(updatedBagId)
-- Scans a specific bag (or all bags if none specified) and 
-- populates/updates self.items, returning a table of all items that
-- have been added/changed/deleted.
    local bagid = updatedBagId or "nil"
    -- self:Print("RebuildInventory: " .. bagid)
    local result = nil    
    local bagType = nil
    
    if updatedBagId ~= nil then
        if self:IsPlayerBagId(updatedBagId) then
            bagType = "player"
        elseif self:IsBankBagId(updatedBagId) then
            bagType = "bank"
        else
            assert(bagType, "Unknown bag type for bagid: " .. updatedBagId)
            -- Support future bag types (guild vault, void storage)
        end

        if bagType ~= nil then -- Scan a specific bag
            local slots = GetContainerNumSlots(updatedBagId)
            if slots then
                result = Sackville:ScanBagSlots(updatedBagId, slots, result)
            else
                self:Print("Bag " .. updatedBagId .. " has no slots")
            end
        end
    else -- Scan everything
        for _, bagid in ipairs(self.bagIndexes.bags) do
            local slots = GetContainerNumSlots(bagid)
            if slots then
                result = Sackville:ScanBagSlots(bagid, slots, result)
            else
                self:Print("Player bag " .. bagid .. " has no slots")
            end
        end
        
        if self.isBankOpen then
            for _, bagid in ipairs(self.bagIndexes.bank) do
                local slots = GetContainerNumSlots(bagid)
                if slots then
                    result = Sackville:ScanBagSlots(bagid, slots, result)
                else
                    self:Print("Bank bag " .. bagid .. " has no slots")
                end
            end
        end
    end
    
    return result
end

-- TODO: Add a description
function Sackville:ScanBagSlots(bagid, slots, priorResult)
    -- self:Print("  ScanBagSlots (" .. bagid .. ", " .. slots .. ")")

    local result = {}
    
    if priorResult then result = priorResult
    else
        result.itemsAdded = {}
        result.itemsChanged = {}
        result.itemsRemoved = {}    
    end
    
    for slot = 1, slots do
        local bagslot = bagid .. '-' ..slot
        local itemid = GetContainerItemID(bagid, slot)
        if itemid then -- slot is not empty
            local isNewItem, itemHasChanged = self:LoadItem(bagid, slot, itemid)
            
            if isNewItem then 
                table.insert(result.itemsAdded, self.items[bagslot])           
            elseif itemHasChanged then
                table.insert(result.itemsChanged, self.items[bagslot])
            end
        else
            if self.items[bagslot] and not self.items[bagslot].removed then
                table.insert(result.itemsRemoved, self.items[bagslot])
            end
        end
    end
    
    -- self:Print(sprintf("  # of items added: %d, changed: %d, removed: %d", 
        -- #result.itemsAdded, #result.itemsChanged, #result.itemsRemoved))
    return result
end

function Sackville:LoadItem(bagid, slot, itemid)
-- Builds an item and assigns it to self.items[bagslot], or updates an existing item

    assert(itemid ~= nil, "LoadItem: nil id at bagslot " .. bagid .. '-' .. slot)


    local isNewItem = nil
    local itemHasChanged = nil    
    local itemCount 
    _, itemCount = GetContainerItemInfo(bagid, slot)

    local savedItem = self.items[bagid .. "-" .. slot]
    -- self:Print(sprintf("LoadItem(%s, %s, %s, savedItem=[%s])", bagid, slot, itemid, self:DumpItem(savedItem)))
    if savedItem == nil then
        isNewItem = true
        itemHasChanged = false
    else
        if savedItem.removed then
            isNewItem = true
            itemHasChanged = true
        elseif savedItem.id == itemid and savedItem.count == itemCount then
            isNewItem = false
            itemHasChanged = false
        else
            isNewItem = false
            itemHasChanged = true
        end
    end
    
    assert(isNewItem ~= nil and itemHasChanged ~= nil, 
        "LoadItem did not properly determine item status")
        
    local itemName, _, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount,
        itemEquipLoc, itemTexture = GetItemInfo(itemid) 
    local isQuestItem, questID, isActive = GetContainerItemQuestInfo(bagid, slot)
    
    local currentItem = savedItem or {}
    
    if isNewItem or itemHasChanged then
        currentItem.id = itemid
        currentItem.count = itemCount
        currentItem.name = itemName or "Unknown item"
        currentItem.quality = itemRarity
        currentItem.level = itemLevel
        currentItem.minLevel = itemMinLevel
        currentItem.equipLoc = itemEquipLoc
        currentItem.type = itemType
        currentItem.subType = itemSubType
        currentItem.stackCount = itemStackCount
        currentItem.texture = itemTexture
        currentItem.bagid = bagid
        currentItem.slot = slot
        currentItem.bagslot = bagid .. "-" .. slot  
        currentItem.isQuestItem = isQuestItem
        currentItem.questID = questID
        currentItem.isActive = isActive
            
        if isNewItem then
            -- self:Print("   Creating new button for " .. self:DumpItem(currentItem))
            currentItem.button = self:createButtonFromItem(currentItem)
        else
            self:UpdateItemButton(currentItem)
        end

        currentItem.removed = false
        self.items[currentItem.bagslot] = currentItem
    end
    
    -- self:Print("Loaded item: " .. self:DumpItem(currentItem))
    return isNewItem, itemHasChanged
end

function Sackville:createButtonFromItem(item)
    -- This wrapper frame allows the calls to self:GetParent():GetID() to
    -- function properly in the inherited Blizzard UI code, since the
    -- parent ID is expected to be the bagid and the child ID is the slot.
    local basename
    local buttonTemplate 
    
    if item.bagid == -1 then
        basename = "SackvilleBankItemButton" .. item.bagslot
        buttonTemplate = "BankItemButtonGenericTemplate"
    else
        basename = "SackvilleItemButton" .. item.bagslot
        buttonTemplate = "ContainerFrameItemButtonTemplate"
    end
    
    wrapper = _G[basename.."_wrapper"]
        or CreateFrame("Frame", basename.."_wrapper", self.dummyFrame)
    wrapper:ClearAllPoints()
    wrapper:SetPoint("TOPLEFT")
    wrapper:SetID(item.bagid)
    wrapper:SetHeight(BUTTON_HEIGHT)
    wrapper:SetWidth(BUTTON_WIDTH)
    wrapper:Show()
    -- 
    item.button = _G[basename] or CreateFrame("Button", basename, wrapper, buttonTemplate)
    item.button:SetAllPoints() -- fill the parent frame
    item.button:SetID(item.slot)
    SetItemButtonTexture(item.button, item.texture)
    SetItemButtonCount(item.button, item.count)
    self:SetQualityOverlay(item)
    self:UpdateButtonCooldown(item)
    item.button:Show()
    -- self:Print("Created button SackvilleItemButton"..item.bagslot..", SetID("..item.slot..")")

    item.button.wrapper = wrapper
    return item.button
end

function Sackville:SetQualityOverlay(item)
-- This will overlay a frame with a colored border over the button.
    local basename = "SackvilleItemButton" .. item.bagslot .. "_quality"
    local button = item.button
    button.qualityFrame = _G[basename]
    or CreateFrame("Frame", basename, item.button)

    
    if (item.quality == nil) or (item.quality < 2) or (item.quality > 7) then 
        button.qualityFrame:Hide()
    else
        local questTexture = (item.questID and not item.isActive) 
        and TEXTURE_ITEM_QUEST_BANG or (item.questID or item.isQuestItem) 
        and TEXTURE_ITEM_QUEST_BORDER
        
        button.qualityFrame.texture = _G[basename .. '_texture']
            or button.qualityFrame:CreateTexture(basename .. '_texture',
            'OVERLAY')
            
        if questTexture then
            button.qualityFrame.texture:SetTexture(questTexture)
        else
            button.qualityFrame.texture:SetTexture(sprintf(
            'Interface\\Addons\\Sackville\\Textures\\UI-Icon-Quality%sBorder.blp',
            item.quality))
        end
        button.qualityFrame.texture:SetBlendMode('BLEND')
        button.qualityFrame.texture:SetAllPoints()
        button.qualityFrame.texture:Show()
        button.qualityFrame:SetAllPoints()
        button.qualityFrame:Show()
    end
end

function Sackville:UpdateButtonCooldown(item, index)
    if item then
        index = index or ""
        assert(item.button ~= nil, "UpdateButtonCooldown " .. index .. ": nil button: " .. self:DumpItem(item))
        
        local cooldown = _G[item.button:GetName().."Cooldown"]; -- FIXME: nil button
        local start, duration, enable = GetContainerItemCooldown(item.bagid, item.slot);
        CooldownFrame_SetTimer(cooldown, start, duration, enable);
        if ( duration > 0 and enable == 0 ) then
            SetItemButtonTextureVertexColor(item.button, 0.4, 0.4, 0.4);
        else
            SetItemButtonTextureVertexColor(item.button, 1, 1, 1);
        end
    end
end

function Sackville:setItemButtonToBlank(item) -- dead code
    -- self:Print("Blanking out button " .. item.bagslot)
    -- item.quality = 0
    -- item.id = 0
    -- item.button:Hide()
    -- item.button:SetID(item.slot) -- blows up 
    -- self:SetQualityOverlay(item)
    -- SetItemButtonTexture(item.button, nil)
    -- SetItemButtonCount(item.button, 0)
end

function Sackville:UpdateItemButton(item)
     self:Print("Updating button for " .. item.name .. " with id " .. item.id .. " at " .. item.bagslot)
    
    if (item.button == nil) then
        self:Print("  nil button, WTF!")
    else
        SetItemButtonTexture(item.button, item.texture)
        SetItemButtonCount(item.button, item.count)
        self:SetQualityOverlay(item)
    end
end

function Sackville:OnInitialize()
-- Code that you want to run when the addon is first loaded goes here.
-- do init tasks here, like loading the Saved Variables, 
-- or setting up slash commands.

    self.db = LibStub("AceDB-3.0"):New("SackvilleDB", defaults, 'Default')
    -- db = self.db.profile
    Sackville:RegisterChatCommand("sack", "ChatCommand")
end

function Sackville:OnEnable()
-- Do more initialization here, that really enables the use of your addon.
-- Register Events, Hook functions, Create Frames, Get information from 
-- the game that wasn't available in OnInitialize
    -- self:RegisterEvent('BAG_UPDATE')
    self:RegisterEvent('PLAYER_ENTERING_WORLD')
    -- self:RegisterEvent("UNIT_INVENTORY_CHANGED")
    -- self:RegisterEvent('ITEM_LOCK_CHANGED')
    -- self:RegisterEvent('BAG_UPDATE_COOLDOWN')
    -- self:RegisterEvent('EQUIPMENT_SETS_CHANGED')
    self:RegisterEvent("BANKFRAME_OPENED")
    self:RegisterEvent("BANKFRAME_CLOSED")
    -- self:RegisterEvent("PLAYER_MONEY")
    -- self:RegisterEvent('AUCTION_HOUSE_SHOW')
    -- self:RegisterEvent('AUCTION_HOUSE_CLOSED')
    -- self:RegisterEvent("CURRENCY_DISPLAY_UPDATE");
    -- self:RegisterEvent("KNOWN_CURRENCY_TYPES_UPDATE");
    self:RegisterEvent("PLAYERBANKSLOTS_CHANGED")
   self:RawHook("ToggleAllBags", "ToggleBags", true)
   self:RawHook("OpenAllBags", "ShowBags", true)
   self:RawHook("OpenBackpack", "ShowBags", true)
   --self:RawHook("CloseBackpack", "HideBags", true)
   --self:RawHook("CloseSpecialWindows", "HideBagsOnEsc", true)    
   -- self:RawHook("SetCurrencyBackpack", "SetCurrencyBackpack", true)
   -- self:RawHookScript(BankFrame, "OnEvent", "BankFrameOnEvent")
   
    self.dummyFrame = CreateFrame("Frame", "SackvilleDummyFrame", UIParent)   
    self:InitializeBags()
end

function Sackville:InitializeBags()
    -- TODO: First clear all existing sections in case this is called a more than once
    for bagIndex=1, #self.bagConfig do
        local bagOptions = self.bagConfig[bagIndex]
        local vbag = Sackville.VirtualBag:New(bagOptions, bagIndex)

        -- for sectionName, sectionOptions in pairs (bagOptions.sections) do        
        for sectionIndex=1, #bagOptions.sections do
            vbag:AddSection(bagOptions.sections[sectionIndex], sectionIndex)
        end
        
        if vbag.type == "player" then
            self.playerBags[bagOptions.name] = vbag
        elseif vbag.type == "bank" then
            self.bankBags[bagOptions.name] = vbag
        else
            -- Support some other bag type
        end
        -- self:Print("Initialized vbag (" .. vbag.name .. ")")
    end
end

function Sackville:OnDisable()
-- Unhook, Unregister Events, Hide frames that you created.
-- You would probably only use an OnDisable if you want to 
-- build a "standby" mode, or be able to toggle modules on/off.
    self:UnregisterEvent('PLAYER_ENTERING_WORLD')
    
    self:HideBags()
end

function Sackville:PLAYER_ENTERING_WORLD()
    self:BuildbagIndexes()
    local result = self:RebuildInventory()

    for _, item in ipairs(result.itemsAdded) do
        local section = self:GetVbagSectionForItem(item, self.playerBags, "player")
        if section then
            section.vbag:AddItem(item, section.name)
        else
            self:Print("Couldn't find a bag for " .. item.name)
        end
    end

    loading = false
end

function Sackville:BAG_UPDATE(event, bagid)
    if (loading) then return end    
    self:Print("BAG_UPDATE: " .. bagid)

    local bagSet
    local bagType

    if self:IsPlayerBagId(bagid) then
        bagSet = self.playerBags
        bagType = "player"

    elseif (self.isBankOpen and self:IsBankBagId(bagid)) then
        bagSet = self.bankBags
        bagType = "bank"
    else
        assert(bagSet, "Unknown bagtype for bagid: " .. bagid)
    end
    -- self:Print("  Chose bagtype " .. bagType)    
    local result
	if bagSet then
        result = self:RebuildInventory(bagid)
    
        for _, item in ipairs(result.itemsAdded) do
            local section = self:GetVbagSectionForItem(item, bagSet, bagType)
            if section then
                section.vbag:AddItem(item, section.name)
            else
                self:Print("Couldn't find a bag for " .. item.name)
            end            
        end
        
        for _, item in ipairs(result.itemsRemoved) do
            item.removed = true
            if item.section.vbag then
                item.section.vbag:RemoveItem(item)
                item.button.wrapper:SetParent(dummyFrame)
            else 
                print("Removed item with no vbag. This shouldn't happen!") 
            end
        end
        
        for _, item in ipairs(result.itemsChanged) do
            item.button.wrapper:Show()
            local section = self:GetVbagSectionForItem(item, bagSet, bagType)
            if section then                
                if (section.vbag.name) ~= item.section.vbag.name or (section.name ~= item.section.name) then
                    item.section.vbag:RemoveItem(item)
                    section.vbag:AddItem(item, section.name)
                end
            else
                self:Print("Couldn't find a bag for " .. item.name)
            end            

        end
    end

    -- No need to refresh the layout immediately if we only removed items.
    -- if result and #result.itemsAdded > 0 or #result.itemsChanged > 0 then
    if result then
        self:LayoutBags()
    end
end

function Sackville:GetVbagSectionForItem(item, bagSet, bagType)
    -- bagSet: A table of virtual bags from which to choose matching section
    -- self:Print("GetVbagSectionForItem: " .. item.name)
    local matchingSections = {}
    
    for bagIndex=1, #self.bagConfig do
        local bagOptions = self.bagConfig[bagIndex]
        if bagType == bagOptions.bagType then
            local vbag = bagSet[bagOptions.name]
            if vbag then
                for sectionIndex=1, #bagOptions.sections do
                    local sectionName = bagOptions.sections[sectionIndex].name
                    assert (sectionName, sprintf("nil sectionName in bagOptions.sections[%d]", sectionIndex))
                    local section = vbag.sections[sectionName]
                    assert(section, sprintf("nil section (%s) in vbag (%s)", 
                        sectionName, vbag.name))
                    if Sackville.Filter:ItemMatchesSection(item, section) then
                        -- self:Print("> Item: " .. item.name ..", found matching section: " .. section.name .. " with prio: " .. section.priority)
                        table.insert(matchingSections, section)
                    end
                end
            else
                self:Print(sprintf("!! nil vbag (%s), item: (%s)",
                    bagOptions.name, item.name))
                assert(vbag)

            end
        end
    end
    
    -- Sort by descending priority. Highest priority wins.
    for i, section in pairs(matchingSections) do
        -- self:Print(sprintf("- Matching section: %s, prio: %d, index: %d",
            -- section.name, section.priority, i))
    end
    local matchComparer = function(a, b) return a.priority > b.priority end
    table.sort(matchingSections, matchComparer)
    
    return matchingSections[1]
end

function Sackville:BANKFRAME_OPENED()
    self:Print("BANKFRAME_OPENED")
    self.isBankOpen = true
    
    if not self.bankInitalized then
        for _, bagid in ipairs(self.bagIndexes.bank) do
            local result = self:RebuildInventory(bagid)           
            for _, item in ipairs(result.itemsAdded) do
                local section = self:GetVbagSectionForItem(item, self.bankBags, "bank")
                if section then
                    section.vbag:AddItem(item, section.name)
                else
                    self:Print("Couldn't find a bag for " .. item.name)
                end
            end
        end
    end
    
    self.bankInitalized = true
    self:ShowBags()
end

function Sackville:BANKFRAME_CLOSED()
    self:Print("BANKFRAME_CLOSED")
    
    if self.isBankOpen then
        for _, vbag in pairs(self.bankBags) do
            vbag:Hide()
        end    
    end    
    
    self.isBankOpen = false

end

function Sackville:PLAYERBANKSLOTS_CHANGED(event, arg1)
    self:Print("PLAYERBANKSLOTS_CHANGED: " .. arg1)
    -- When (arg1 <= NUM_BANKGENERIC_SLOTS), arg1 is the index of the generic
    -- bank slot that changed. When (arg1 > NUM_BANKGENERIC_SLOTS), 
    -- (arg1 - NUM_BANKGENERIC_SLOTS) is the index of the equipped bank bag that changed. 
    local bagid
    if (arg1 <= NUM_BANKGENERIC_SLOTS) then -- generic bank slot changed
        self:BAG_UPDATE(event, -1)
    else
        -- bagid = arg1 - NUM_BANKGENERIC_SLOTS    
        -- self:BAG_UPDATE(event, bagid)
    end
end

function Sackville:BAG_UPDATE_COOLDOWN(event, cid)
    -- cid = cid or "?"
    self:Print("BAG_UPDATE_COOLDOWN:" .. (cid or "?"))
    
    for i, item in pairs(self.items) do
        if item ~= nil then self:UpdateButtonCooldown(item, i) end
    end

end

function Sackville:UNIT_INVENTORY_CHANGED(e, unitid)
    if (loading) then return end
    self:Print("UNIT_INVENTORY_CHANGED, unitid:" .. unitid)
end

function Sackville:DISPLAY_SIZE_CHANGED()
    self:Print("DISPLAY_SIZE_CHANGED")
end

function Sackville:ITEM_LOCK_CHANGED(event, bagid, slot)
-- Desaturate item button when it is locked (probably because you're moving it)
    -- self:Print("ITEM_LOCK_CHANGED")
    if not (bagid and slot) then return end
    local item = self.items[bagid .. '-' .. slot]
    if (item ~= nil) then
        local _, _, locked = GetContainerItemInfo(bagid, slot);
        SetItemButtonDesaturated(item.button, locked);
    end

end

function Sackville:ChatCommand(input)
    input = (input) and input or ""
    -- self:Print("ChatCommand(" .. input .. ")")
    if not input or input:trim() == '' then
        self:ToggleBags()
    elseif input == 'bi' then
        self:RebuildInventory()
    elseif input == 'lay' then
        self:LayoutBags()
    end
end

function Sackville:ToggleBags()
    if showBags then
        self:HideBags()
    else
        self:ShowBags()
    end
end

function Sackville:ShowBags()
    -- self:Print("ShowBags()")
    showBags = true
    Sackville:RegisterEvent("BAG_UPDATE")
	Sackville:RegisterEvent("ITEM_LOCK_CHANGED")
	Sackville:RegisterEvent("BAG_UPDATE_COOLDOWN")
	Sackville:RegisterEvent("DISPLAY_SIZE_CHANGED")
    
    for _, vbag in pairs(self.playerBags) do
        vbag:Show()
    end
    
    if self.isBankOpen then
        for _, vbag in pairs(self.bankBags) do
            vbag:Show()
        end    
    end
    PlaySound("igBackPackOpen")

    self:LayoutBags()
end

function Sackville:HideBags()
    -- self:Print("HideBags()")
    showBags = false
	Sackville:UnregisterEvent("BAG_UPDATE")
	Sackville:UnregisterEvent("ITEM_LOCK_CHANGED")
	Sackville:UnregisterEvent("BAG_UPDATE_COOLDOWN")
	Sackville:UnregisterEvent("DISPLAY_SIZE_CHANGED")
    
    for _, vbag in pairs(self.playerBags) do
        vbag:Hide()
    end
    
    if self.isBankOpen then
        for _, vbag in pairs(self.bankBags) do
            vbag:Hide()
        end    
    end    
    PlaySound("igBackPackClose")    
end

function Sackville:LayoutBags()
    -- self:Print("LayoutBags()")

    local positionState = {
        x_offset = -20,
        y_offset = 80,
        widest_bag = 0
    }
    
    -- FIXME: Make sure we display in the proper order
    for bagName, vbag in pairs(self.playerBags) do
        vbag:LayoutItems()
        positionState = PositionBag(vbag, positionState)
    end
    
    if self.isBankOpen then
        for bagName, vbag in pairs(self.bankBags) do
            vbag:LayoutItems()
            positionState = PositionBag(vbag, positionState)
        end
    end
end


